# coding: utf-8

# flake8: noqa

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.8.1-develop.1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

# import apis into sdk package
from curia.api.swagger_client.api.cohorts_api import CohortsApi
from curia.api.swagger_client.api.data_queries_api import DataQueriesApi
from curia.api.swagger_client.api.data_stores_api import DataStoresApi
from curia.api.swagger_client.api.data_tables_api import DataTablesApi
from curia.api.swagger_client.api.databases_api import DatabasesApi
from curia.api.swagger_client.api.dataset_columns_api import DatasetColumnsApi
from curia.api.swagger_client.api.datasets_api import DatasetsApi
from curia.api.swagger_client.api.feature_stores_api import FeatureStoresApi
from curia.api.swagger_client.api.features_api import FeaturesApi
from curia.api.swagger_client.api.internal_api import InternalApi
from curia.api.swagger_client.api.model_job_outputs_api import ModelJobOutputsApi
from curia.api.swagger_client.api.model_job_statuses_api import ModelJobStatusesApi
from curia.api.swagger_client.api.model_jobs_api import ModelJobsApi
from curia.api.swagger_client.api.model_populations_api import ModelPopulationsApi
from curia.api.swagger_client.api.models_api import ModelsApi
from curia.api.swagger_client.api.organization_settings_api import OrganizationSettingsApi
from curia.api.swagger_client.api.organizations_api import OrganizationsApi
from curia.api.swagger_client.api.platform_api import PlatformApi
from curia.api.swagger_client.api.population_stores_api import PopulationStoresApi
from curia.api.swagger_client.api.project_members_api import ProjectMembersApi
from curia.api.swagger_client.api.projects_api import ProjectsApi
from curia.api.swagger_client.api.scheduler_api import SchedulerApi
from curia.api.swagger_client.api.tasks_api import TasksApi
from curia.api.swagger_client.api.user_audit_trail_api import UserAuditTrailApi
from curia.api.swagger_client.api.user_favorites_api import UserFavoritesApi
from curia.api.swagger_client.api.user_management_api import UserManagementApi
from curia.api.swagger_client.api.workflows_api import WorkflowsApi
# import ApiClient
from curia.api.swagger_client.api_client import ApiClient
from curia.api.swagger_client.configuration import Configuration
# import models into sdk package
from curia.api.swagger_client.models.all_of_data_table_dataset import AllOfDataTableDataset
from curia.api.swagger_client.models.arithmetic_expression import ArithmeticExpression
from curia.api.swagger_client.models.body import Body
from curia.api.swagger_client.models.body1 import Body1
from curia.api.swagger_client.models.body2 import Body2
from curia.api.swagger_client.models.body3 import Body3
from curia.api.swagger_client.models.body4 import Body4
from curia.api.swagger_client.models.body5 import Body5
from curia.api.swagger_client.models.boolean_expression import BooleanExpression
from curia.api.swagger_client.models.cohort import Cohort
from curia.api.swagger_client.models.cohort_definition import CohortDefinition
from curia.api.swagger_client.models.cohort_filter import CohortFilter
from curia.api.swagger_client.models.cohort_filter_condition import CohortFilterCondition
from curia.api.swagger_client.models.cohort_joined_cohort_response_dto import CohortJoinedCohortResponseDto
from curia.api.swagger_client.models.cohort_joined_model_job_response_dto import CohortJoinedModelJobResponseDto
from curia.api.swagger_client.models.cohort_response_dto import CohortResponseDto
from curia.api.swagger_client.models.cohort_results import CohortResults
from curia.api.swagger_client.models.cohort_window import CohortWindow
from curia.api.swagger_client.models.condition import Condition
from curia.api.swagger_client.models.container_config import ContainerConfig
from curia.api.swagger_client.models.create_cohort_dto import CreateCohortDto
from curia.api.swagger_client.models.create_data_query_dto import CreateDataQueryDto
from curia.api.swagger_client.models.create_data_store_dto import CreateDataStoreDto
from curia.api.swagger_client.models.create_data_table_dto import CreateDataTableDto
from curia.api.swagger_client.models.create_database_dto import CreateDatabaseDto
from curia.api.swagger_client.models.create_dataset_column_dto import CreateDatasetColumnDto
from curia.api.swagger_client.models.create_dataset_dto import CreateDatasetDto
from curia.api.swagger_client.models.create_feature_dto import CreateFeatureDto
from curia.api.swagger_client.models.create_feature_store_dto import CreateFeatureStoreDto
from curia.api.swagger_client.models.create_many_cohort_dto import CreateManyCohortDto
from curia.api.swagger_client.models.create_many_data_query_dto import CreateManyDataQueryDto
from curia.api.swagger_client.models.create_many_data_store_dto import CreateManyDataStoreDto
from curia.api.swagger_client.models.create_many_dataset_column_dto import CreateManyDatasetColumnDto
from curia.api.swagger_client.models.create_many_dataset_dto import CreateManyDatasetDto
from curia.api.swagger_client.models.create_many_feature_dto import CreateManyFeatureDto
from curia.api.swagger_client.models.create_many_feature_store_dto import CreateManyFeatureStoreDto
from curia.api.swagger_client.models.create_many_model_dto import CreateManyModelDto
from curia.api.swagger_client.models.create_many_model_job_dto import CreateManyModelJobDto
from curia.api.swagger_client.models.create_many_model_job_output_dto import CreateManyModelJobOutputDto
from curia.api.swagger_client.models.create_many_model_population_dto import CreateManyModelPopulationDto
from curia.api.swagger_client.models.create_many_organization_dto import CreateManyOrganizationDto
from curia.api.swagger_client.models.create_many_organization_setting_dto import CreateManyOrganizationSettingDto
from curia.api.swagger_client.models.create_many_population_store_dto import CreateManyPopulationStoreDto
from curia.api.swagger_client.models.create_many_project_dto import CreateManyProjectDto
from curia.api.swagger_client.models.create_many_project_member_dto import CreateManyProjectMemberDto
from curia.api.swagger_client.models.create_many_scheduler_dto import CreateManySchedulerDto
from curia.api.swagger_client.models.create_many_task_dto import CreateManyTaskDto
from curia.api.swagger_client.models.create_many_task_execution_dto import CreateManyTaskExecutionDto
from curia.api.swagger_client.models.create_many_task_execution_status_dto import CreateManyTaskExecutionStatusDto
from curia.api.swagger_client.models.create_many_user_favorite_dto import CreateManyUserFavoriteDto
from curia.api.swagger_client.models.create_many_workflow_dto import CreateManyWorkflowDto
from curia.api.swagger_client.models.create_many_workflow_execution_dto import CreateManyWorkflowExecutionDto
from curia.api.swagger_client.models.create_many_workflow_execution_spec_dto import CreateManyWorkflowExecutionSpecDto
from curia.api.swagger_client.models.create_many_workflow_template_dto import CreateManyWorkflowTemplateDto
from curia.api.swagger_client.models.create_model_dto import CreateModelDto
from curia.api.swagger_client.models.create_model_job_dto import CreateModelJobDto
from curia.api.swagger_client.models.create_model_job_output_dto import CreateModelJobOutputDto
from curia.api.swagger_client.models.create_model_job_status_dto import CreateModelJobStatusDto
from curia.api.swagger_client.models.create_model_population_dto import CreateModelPopulationDto
from curia.api.swagger_client.models.create_organization_dto import CreateOrganizationDto
from curia.api.swagger_client.models.create_organization_setting_dto import CreateOrganizationSettingDto
from curia.api.swagger_client.models.create_population_store_dto import CreatePopulationStoreDto
from curia.api.swagger_client.models.create_project_dto import CreateProjectDto
from curia.api.swagger_client.models.create_project_member_dto import CreateProjectMemberDto
from curia.api.swagger_client.models.create_scheduler_dto import CreateSchedulerDto
from curia.api.swagger_client.models.create_task_dto import CreateTaskDto
from curia.api.swagger_client.models.create_task_execution_dto import CreateTaskExecutionDto
from curia.api.swagger_client.models.create_task_execution_status_dto import CreateTaskExecutionStatusDto
from curia.api.swagger_client.models.create_user_audit_trail_dto import CreateUserAuditTrailDto
from curia.api.swagger_client.models.create_user_favorite_dto import CreateUserFavoriteDto
from curia.api.swagger_client.models.create_workflow_dto import CreateWorkflowDto
from curia.api.swagger_client.models.create_workflow_execution_dto import CreateWorkflowExecutionDto
from curia.api.swagger_client.models.create_workflow_execution_spec_dto import CreateWorkflowExecutionSpecDto
from curia.api.swagger_client.models.create_workflow_execution_status_dto import CreateWorkflowExecutionStatusDto
from curia.api.swagger_client.models.create_workflow_template_dto import CreateWorkflowTemplateDto
from curia.api.swagger_client.models.data_query import DataQuery
from curia.api.swagger_client.models.data_query_joined_dataset_response_dto import DataQueryJoinedDatasetResponseDto
from curia.api.swagger_client.models.data_query_response_dto import DataQueryResponseDto
from curia.api.swagger_client.models.data_store import DataStore
from curia.api.swagger_client.models.data_store_response_dto import DataStoreResponseDto
from curia.api.swagger_client.models.data_table import DataTable
from curia.api.swagger_client.models.data_table_joined_database_response_dto import DataTableJoinedDatabaseResponseDto
from curia.api.swagger_client.models.data_table_joined_dataset_response_dto import DataTableJoinedDatasetResponseDto
from curia.api.swagger_client.models.data_table_response_dto import DataTableResponseDto
from curia.api.swagger_client.models.database import Database
from curia.api.swagger_client.models.database_joined_data_table_response_dto import DatabaseJoinedDataTableResponseDto
from curia.api.swagger_client.models.database_response_dto import DatabaseResponseDto
from curia.api.swagger_client.models.dataset import Dataset
from curia.api.swagger_client.models.dataset_column import DatasetColumn
from curia.api.swagger_client.models.dataset_column_joined_dataset_response_dto import DatasetColumnJoinedDatasetResponseDto
from curia.api.swagger_client.models.dataset_column_response_dto import DatasetColumnResponseDto
from curia.api.swagger_client.models.dataset_joined_data_query_response_dto import DatasetJoinedDataQueryResponseDto
from curia.api.swagger_client.models.dataset_joined_dataset_column_response_dto import DatasetJoinedDatasetColumnResponseDto
from curia.api.swagger_client.models.dataset_response_dto import DatasetResponseDto
from curia.api.swagger_client.models.datasetsuploadlargecomplete_parts import DatasetsuploadlargecompleteParts
from curia.api.swagger_client.models.feature import Feature
from curia.api.swagger_client.models.feature_category import FeatureCategory
from curia.api.swagger_client.models.feature_joined_feature_sub_category_joined_feature_category_response_dto import FeatureJoinedFeatureSubCategoryJoinedFeatureCategoryResponseDto
from curia.api.swagger_client.models.feature_joined_feature_sub_category_response_dto import FeatureJoinedFeatureSubCategoryResponseDto
from curia.api.swagger_client.models.feature_joined_model_population_response_dto import FeatureJoinedModelPopulationResponseDto
from curia.api.swagger_client.models.feature_joined_organization_feature_exclusion_response_dto import FeatureJoinedOrganizationFeatureExclusionResponseDto
from curia.api.swagger_client.models.feature_response_dto import FeatureResponseDto
from curia.api.swagger_client.models.feature_store import FeatureStore
from curia.api.swagger_client.models.feature_store_response_dto import FeatureStoreResponseDto
from curia.api.swagger_client.models.feature_sub_category import FeatureSubCategory
from curia.api.swagger_client.models.geographic_counts import GeographicCounts
from curia.api.swagger_client.models.geographic_queries import GeographicQueries
from curia.api.swagger_client.models.get_many_cohort_response_dto import GetManyCohortResponseDto
from curia.api.swagger_client.models.get_many_data_query_response_dto import GetManyDataQueryResponseDto
from curia.api.swagger_client.models.get_many_data_store_response_dto import GetManyDataStoreResponseDto
from curia.api.swagger_client.models.get_many_data_table_response_dto import GetManyDataTableResponseDto
from curia.api.swagger_client.models.get_many_database_response_dto import GetManyDatabaseResponseDto
from curia.api.swagger_client.models.get_many_dataset_column_response_dto import GetManyDatasetColumnResponseDto
from curia.api.swagger_client.models.get_many_dataset_response_dto import GetManyDatasetResponseDto
from curia.api.swagger_client.models.get_many_feature_response_dto import GetManyFeatureResponseDto
from curia.api.swagger_client.models.get_many_feature_store_response_dto import GetManyFeatureStoreResponseDto
from curia.api.swagger_client.models.get_many_model_job_output_response_dto import GetManyModelJobOutputResponseDto
from curia.api.swagger_client.models.get_many_model_job_response_dto import GetManyModelJobResponseDto
from curia.api.swagger_client.models.get_many_model_job_status_response_dto import GetManyModelJobStatusResponseDto
from curia.api.swagger_client.models.get_many_model_population_response_dto import GetManyModelPopulationResponseDto
from curia.api.swagger_client.models.get_many_model_response_dto import GetManyModelResponseDto
from curia.api.swagger_client.models.get_many_organization_response_dto import GetManyOrganizationResponseDto
from curia.api.swagger_client.models.get_many_organization_setting_response_dto import GetManyOrganizationSettingResponseDto
from curia.api.swagger_client.models.get_many_population_store_response_dto import GetManyPopulationStoreResponseDto
from curia.api.swagger_client.models.get_many_project_member_response_dto import GetManyProjectMemberResponseDto
from curia.api.swagger_client.models.get_many_project_response_dto import GetManyProjectResponseDto
from curia.api.swagger_client.models.get_many_scheduler_response_dto import GetManySchedulerResponseDto
from curia.api.swagger_client.models.get_many_task_execution_response_dto import GetManyTaskExecutionResponseDto
from curia.api.swagger_client.models.get_many_task_execution_status_response_dto import GetManyTaskExecutionStatusResponseDto
from curia.api.swagger_client.models.get_many_task_response_dto import GetManyTaskResponseDto
from curia.api.swagger_client.models.get_many_user_audit_trail_response_dto import GetManyUserAuditTrailResponseDto
from curia.api.swagger_client.models.get_many_user_favorite_response_dto import GetManyUserFavoriteResponseDto
from curia.api.swagger_client.models.get_many_workflow_execution_response_dto import GetManyWorkflowExecutionResponseDto
from curia.api.swagger_client.models.get_many_workflow_execution_spec_response_dto import GetManyWorkflowExecutionSpecResponseDto
from curia.api.swagger_client.models.get_many_workflow_execution_status_response_dto import GetManyWorkflowExecutionStatusResponseDto
from curia.api.swagger_client.models.get_many_workflow_response_dto import GetManyWorkflowResponseDto
from curia.api.swagger_client.models.get_many_workflow_template_response_dto import GetManyWorkflowTemplateResponseDto
from curia.api.swagger_client.models.intervention_definition import InterventionDefinition
from curia.api.swagger_client.models.json import Json
from curia.api.swagger_client.models.model import Model
from curia.api.swagger_client.models.model_job import ModelJob
from curia.api.swagger_client.models.model_job_config import ModelJobConfig
from curia.api.swagger_client.models.model_job_event import ModelJobEvent
from curia.api.swagger_client.models.model_job_joined_cohort_response_dto import ModelJobJoinedCohortResponseDto
from curia.api.swagger_client.models.model_job_joined_data_query_response_dto import ModelJobJoinedDataQueryResponseDto
from curia.api.swagger_client.models.model_job_joined_dataset_response_dto import ModelJobJoinedDatasetResponseDto
from curia.api.swagger_client.models.model_job_joined_feature_sub_category_response_dto import ModelJobJoinedFeatureSubCategoryResponseDto
from curia.api.swagger_client.models.model_job_joined_model_job_output_response_dto import ModelJobJoinedModelJobOutputResponseDto
from curia.api.swagger_client.models.model_job_joined_model_job_status_response_dto import ModelJobJoinedModelJobStatusResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_cohort_definition_joined_period_definition_response_dto import ModelJobJoinedModelPopulationJoinedCohortDefinitionJoinedPeriodDefinitionResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_cohort_definition_response_dto import ModelJobJoinedModelPopulationJoinedCohortDefinitionResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_data_query_joined_dataset_response_dto import ModelJobJoinedModelPopulationJoinedDataQueryJoinedDatasetResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_data_query_response_dto import ModelJobJoinedModelPopulationJoinedDataQueryResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_intervention_definition_response_dto import ModelJobJoinedModelPopulationJoinedInterventionDefinitionResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_joined_outcome_definition_response_dto import ModelJobJoinedModelPopulationJoinedOutcomeDefinitionResponseDto
from curia.api.swagger_client.models.model_job_joined_model_population_response_dto import ModelJobJoinedModelPopulationResponseDto
from curia.api.swagger_client.models.model_job_joined_model_response_dto import ModelJobJoinedModelResponseDto
from curia.api.swagger_client.models.model_job_joined_project_response_dto import ModelJobJoinedProjectResponseDto
from curia.api.swagger_client.models.model_job_output import ModelJobOutput
from curia.api.swagger_client.models.model_job_output_joined_dataset_response_dto import ModelJobOutputJoinedDatasetResponseDto
from curia.api.swagger_client.models.model_job_output_joined_model_job_joined_project_response_dto import ModelJobOutputJoinedModelJobJoinedProjectResponseDto
from curia.api.swagger_client.models.model_job_output_joined_model_job_response_dto import ModelJobOutputJoinedModelJobResponseDto
from curia.api.swagger_client.models.model_job_output_response_dto import ModelJobOutputResponseDto
from curia.api.swagger_client.models.model_job_response_dto import ModelJobResponseDto
from curia.api.swagger_client.models.model_job_status import ModelJobStatus
from curia.api.swagger_client.models.model_job_status_joined_model_job_joined_project_response_dto import ModelJobStatusJoinedModelJobJoinedProjectResponseDto
from curia.api.swagger_client.models.model_job_status_joined_model_job_response_dto import ModelJobStatusJoinedModelJobResponseDto
from curia.api.swagger_client.models.model_job_status_response_dto import ModelJobStatusResponseDto
from curia.api.swagger_client.models.model_joined_cohort_joined_model_job_response_dto import ModelJoinedCohortJoinedModelJobResponseDto
from curia.api.swagger_client.models.model_joined_cohort_response_dto import ModelJoinedCohortResponseDto
from curia.api.swagger_client.models.model_joined_model_job_joined_cohort_response_dto import ModelJoinedModelJobJoinedCohortResponseDto
from curia.api.swagger_client.models.model_joined_model_job_joined_dataset_response_dto import ModelJoinedModelJobJoinedDatasetResponseDto
from curia.api.swagger_client.models.model_joined_model_job_joined_feature_sub_category_response_dto import ModelJoinedModelJobJoinedFeatureSubCategoryResponseDto
from curia.api.swagger_client.models.model_joined_model_job_response_dto import ModelJoinedModelJobResponseDto
from curia.api.swagger_client.models.model_joined_model_output_details_response_dto import ModelJoinedModelOutputDetailsResponseDto
from curia.api.swagger_client.models.model_joined_project_response_dto import ModelJoinedProjectResponseDto
from curia.api.swagger_client.models.model_joined_user_favorite_response_dto import ModelJoinedUserFavoriteResponseDto
from curia.api.swagger_client.models.model_output_details import ModelOutputDetails
from curia.api.swagger_client.models.model_population import ModelPopulation
from curia.api.swagger_client.models.model_population_joined_cohort_definition_joined_period_definition_response_dto import ModelPopulationJoinedCohortDefinitionJoinedPeriodDefinitionResponseDto
from curia.api.swagger_client.models.model_population_joined_cohort_definition_response_dto import ModelPopulationJoinedCohortDefinitionResponseDto
from curia.api.swagger_client.models.model_population_joined_cohort_results_response_dto import ModelPopulationJoinedCohortResultsResponseDto
from curia.api.swagger_client.models.model_population_joined_data_query_joined_dataset_response_dto import ModelPopulationJoinedDataQueryJoinedDatasetResponseDto
from curia.api.swagger_client.models.model_population_joined_data_query_response_dto import ModelPopulationJoinedDataQueryResponseDto
from curia.api.swagger_client.models.model_population_joined_feature_response_dto import ModelPopulationJoinedFeatureResponseDto
from curia.api.swagger_client.models.model_population_joined_intervention_definition_response_dto import ModelPopulationJoinedInterventionDefinitionResponseDto
from curia.api.swagger_client.models.model_population_joined_model_job_joined_model_response_dto import ModelPopulationJoinedModelJobJoinedModelResponseDto
from curia.api.swagger_client.models.model_population_joined_model_job_response_dto import ModelPopulationJoinedModelJobResponseDto
from curia.api.swagger_client.models.model_population_joined_outcome_definition_response_dto import ModelPopulationJoinedOutcomeDefinitionResponseDto
from curia.api.swagger_client.models.model_population_response_dto import ModelPopulationResponseDto
from curia.api.swagger_client.models.model_response_dto import ModelResponseDto
from curia.api.swagger_client.models.object import Object
from curia.api.swagger_client.models.organization import Organization
from curia.api.swagger_client.models.organization_feature_category_exclusion import OrganizationFeatureCategoryExclusion
from curia.api.swagger_client.models.organization_feature_exclusion import OrganizationFeatureExclusion
from curia.api.swagger_client.models.organization_response_dto import OrganizationResponseDto
from curia.api.swagger_client.models.organization_setting import OrganizationSetting
from curia.api.swagger_client.models.organization_setting_response_dto import OrganizationSettingResponseDto
from curia.api.swagger_client.models.outcome_definition import OutcomeDefinition
from curia.api.swagger_client.models.period_definition import PeriodDefinition
from curia.api.swagger_client.models.person_set import PersonSet
from curia.api.swagger_client.models.pipeline_config import PipelineConfig
from curia.api.swagger_client.models.population import Population
from curia.api.swagger_client.models.population_store import PopulationStore
from curia.api.swagger_client.models.population_store_response_dto import PopulationStoreResponseDto
from curia.api.swagger_client.models.project import Project
from curia.api.swagger_client.models.project_joined_model_joined_cohort_response_dto import ProjectJoinedModelJoinedCohortResponseDto
from curia.api.swagger_client.models.project_joined_model_response_dto import ProjectJoinedModelResponseDto
from curia.api.swagger_client.models.project_joined_project_response_dto import ProjectJoinedProjectResponseDto
from curia.api.swagger_client.models.project_joined_user_favorite_response_dto import ProjectJoinedUserFavoriteResponseDto
from curia.api.swagger_client.models.project_member import ProjectMember
from curia.api.swagger_client.models.project_member_joined_project_response_dto import ProjectMemberJoinedProjectResponseDto
from curia.api.swagger_client.models.project_member_response_dto import ProjectMemberResponseDto
from curia.api.swagger_client.models.project_response_dto import ProjectResponseDto
from curia.api.swagger_client.models.scheduler import Scheduler
from curia.api.swagger_client.models.scheduler_joined_workflow_response_dto import SchedulerJoinedWorkflowResponseDto
from curia.api.swagger_client.models.scheduler_response_dto import SchedulerResponseDto
from curia.api.swagger_client.models.select_expression import SelectExpression
from curia.api.swagger_client.models.state_data import StateData
from curia.api.swagger_client.models.task import Task
from curia.api.swagger_client.models.task_execution import TaskExecution
from curia.api.swagger_client.models.task_execution_joined_task_response_dto import TaskExecutionJoinedTaskResponseDto
from curia.api.swagger_client.models.task_execution_joined_workflow_execution_response_dto import TaskExecutionJoinedWorkflowExecutionResponseDto
from curia.api.swagger_client.models.task_execution_response_dto import TaskExecutionResponseDto
from curia.api.swagger_client.models.task_execution_status import TaskExecutionStatus
from curia.api.swagger_client.models.task_execution_status_joined_task_execution_joined_workflow_execution_response_dto import TaskExecutionStatusJoinedTaskExecutionJoinedWorkflowExecutionResponseDto
from curia.api.swagger_client.models.task_execution_status_joined_task_execution_response_dto import TaskExecutionStatusJoinedTaskExecutionResponseDto
from curia.api.swagger_client.models.task_execution_status_response_dto import TaskExecutionStatusResponseDto
from curia.api.swagger_client.models.task_inputs import TaskInputs
from curia.api.swagger_client.models.task_outputs import TaskOutputs
from curia.api.swagger_client.models.task_response_dto import TaskResponseDto
from curia.api.swagger_client.models.update_cohort_dto import UpdateCohortDto
from curia.api.swagger_client.models.update_data_query_dto import UpdateDataQueryDto
from curia.api.swagger_client.models.update_data_store_dto import UpdateDataStoreDto
from curia.api.swagger_client.models.update_data_table_dto import UpdateDataTableDto
from curia.api.swagger_client.models.update_database_dto import UpdateDatabaseDto
from curia.api.swagger_client.models.update_dataset_column_dto import UpdateDatasetColumnDto
from curia.api.swagger_client.models.update_dataset_dto import UpdateDatasetDto
from curia.api.swagger_client.models.update_feature_dto import UpdateFeatureDto
from curia.api.swagger_client.models.update_feature_store_dto import UpdateFeatureStoreDto
from curia.api.swagger_client.models.update_model_dto import UpdateModelDto
from curia.api.swagger_client.models.update_model_job_dto import UpdateModelJobDto
from curia.api.swagger_client.models.update_model_job_output_dto import UpdateModelJobOutputDto
from curia.api.swagger_client.models.update_model_job_status_dto import UpdateModelJobStatusDto
from curia.api.swagger_client.models.update_model_population_dto import UpdateModelPopulationDto
from curia.api.swagger_client.models.update_organization_dto import UpdateOrganizationDto
from curia.api.swagger_client.models.update_organization_setting_dto import UpdateOrganizationSettingDto
from curia.api.swagger_client.models.update_population_store_dto import UpdatePopulationStoreDto
from curia.api.swagger_client.models.update_project_dto import UpdateProjectDto
from curia.api.swagger_client.models.update_project_member_dto import UpdateProjectMemberDto
from curia.api.swagger_client.models.update_scheduler_dto import UpdateSchedulerDto
from curia.api.swagger_client.models.update_task_dto import UpdateTaskDto
from curia.api.swagger_client.models.update_task_execution_dto import UpdateTaskExecutionDto
from curia.api.swagger_client.models.update_task_execution_status_dto import UpdateTaskExecutionStatusDto
from curia.api.swagger_client.models.update_user_audit_trail_dto import UpdateUserAuditTrailDto
from curia.api.swagger_client.models.update_user_dto import UpdateUserDto
from curia.api.swagger_client.models.update_user_favorite_dto import UpdateUserFavoriteDto
from curia.api.swagger_client.models.update_workflow_dto import UpdateWorkflowDto
from curia.api.swagger_client.models.update_workflow_execution_dto import UpdateWorkflowExecutionDto
from curia.api.swagger_client.models.update_workflow_execution_spec_dto import UpdateWorkflowExecutionSpecDto
from curia.api.swagger_client.models.update_workflow_execution_status_dto import UpdateWorkflowExecutionStatusDto
from curia.api.swagger_client.models.update_workflow_template_dto import UpdateWorkflowTemplateDto
from curia.api.swagger_client.models.user_audit_trail_response_dto import UserAuditTrailResponseDto
from curia.api.swagger_client.models.user_favorite import UserFavorite
from curia.api.swagger_client.models.user_favorite_joined_model_joined_project_response_dto import UserFavoriteJoinedModelJoinedProjectResponseDto
from curia.api.swagger_client.models.user_favorite_joined_model_response_dto import UserFavoriteJoinedModelResponseDto
from curia.api.swagger_client.models.user_favorite_response_dto import UserFavoriteResponseDto
from curia.api.swagger_client.models.workflow import Workflow
from curia.api.swagger_client.models.workflow_execution import WorkflowExecution
from curia.api.swagger_client.models.workflow_execution_response_dto import WorkflowExecutionResponseDto
from curia.api.swagger_client.models.workflow_execution_spec import WorkflowExecutionSpec
from curia.api.swagger_client.models.workflow_execution_spec_joined_workflow_execution_response_dto import WorkflowExecutionSpecJoinedWorkflowExecutionResponseDto
from curia.api.swagger_client.models.workflow_execution_spec_response_dto import WorkflowExecutionSpecResponseDto
from curia.api.swagger_client.models.workflow_execution_status import WorkflowExecutionStatus
from curia.api.swagger_client.models.workflow_execution_status_joined_workflow_execution_response_dto import WorkflowExecutionStatusJoinedWorkflowExecutionResponseDto
from curia.api.swagger_client.models.workflow_execution_status_response_dto import WorkflowExecutionStatusResponseDto
from curia.api.swagger_client.models.workflow_joined_scheduler_response_dto import WorkflowJoinedSchedulerResponseDto
from curia.api.swagger_client.models.workflow_joined_workflow_execution_response_dto import WorkflowJoinedWorkflowExecutionResponseDto
from curia.api.swagger_client.models.workflow_joined_workflow_template_response_dto import WorkflowJoinedWorkflowTemplateResponseDto
from curia.api.swagger_client.models.workflow_response_dto import WorkflowResponseDto
from curia.api.swagger_client.models.workflow_template import WorkflowTemplate
from curia.api.swagger_client.models.workflow_template_definition import WorkflowTemplateDefinition
from curia.api.swagger_client.models.workflow_template_parameters import WorkflowTemplateParameters
from curia.api.swagger_client.models.workflow_template_response_dto import WorkflowTemplateResponseDto
from curia.api.swagger_client.models.zip_data import ZipData



# coding: utf-8

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.8.1-develop.1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class Scheduler(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'id': 'str',
        'last_updated_by': 'str',
        'created_by': 'str',
        'created_at': 'datetime',
        'updated_at': 'datetime',
        'archived_at': 'datetime',
        'version': 'float',
        'type': 'str',
        'workflow_id': 'str',
        'schedule': 'str',
        'parameters': 'WorkflowTemplateParameters',
        'organization_id': 'str',
        'job_key': 'str',
        'last_run_at': 'datetime',
        'next_run_at': 'datetime',
        'is_paused': 'object',
        'organization': 'Organization',
        'workflow': 'Workflow'
    }

    attribute_map = {
        'id': 'id',
        'last_updated_by': 'lastUpdatedBy',
        'created_by': 'createdBy',
        'created_at': 'createdAt',
        'updated_at': 'updatedAt',
        'archived_at': 'archivedAt',
        'version': 'version',
        'type': 'type',
        'workflow_id': 'workflowId',
        'schedule': 'schedule',
        'parameters': 'parameters',
        'organization_id': 'organizationId',
        'job_key': 'jobKey',
        'last_run_at': 'lastRunAt',
        'next_run_at': 'nextRunAt',
        'is_paused': 'isPaused',
        'organization': 'organization',
        'workflow': 'workflow'
    }

    def __init__(self, id=None, last_updated_by=None, created_by=None, created_at=None, updated_at=None, archived_at=None, version=None, type=None, workflow_id=None, schedule=None, parameters=None, organization_id=None, job_key=None, last_run_at=None, next_run_at=None, is_paused=None, organization=None, workflow=None):  # noqa: E501
        """Scheduler - a model defined in Swagger"""  # noqa: E501
        self._id = None
        self._last_updated_by = None
        self._created_by = None
        self._created_at = None
        self._updated_at = None
        self._archived_at = None
        self._version = None
        self._type = None
        self._workflow_id = None
        self._schedule = None
        self._parameters = None
        self._organization_id = None
        self._job_key = None
        self._last_run_at = None
        self._next_run_at = None
        self._is_paused = None
        self._organization = None
        self._workflow = None
        self.discriminator = None
        if id is not None:
            self.id = id
        if last_updated_by is not None:
            self.last_updated_by = last_updated_by
        if created_by is not None:
            self.created_by = created_by
        if created_at is not None:
            self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at
        if archived_at is not None:
            self.archived_at = archived_at
        if version is not None:
            self.version = version
        self.type = type
        self.workflow_id = workflow_id
        self.schedule = schedule
        if parameters is not None:
            self.parameters = parameters
        self.organization_id = organization_id
        if job_key is not None:
            self.job_key = job_key
        if last_run_at is not None:
            self.last_run_at = last_run_at
        if next_run_at is not None:
            self.next_run_at = next_run_at
        self.is_paused = is_paused
        if organization is not None:
            self.organization = organization
        if workflow is not None:
            self.workflow = workflow

    @property
    def id(self):
        """Gets the id of this Scheduler.  # noqa: E501


        :return: The id of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Scheduler.


        :param id: The id of this Scheduler.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def last_updated_by(self):
        """Gets the last_updated_by of this Scheduler.  # noqa: E501


        :return: The last_updated_by of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, last_updated_by):
        """Sets the last_updated_by of this Scheduler.


        :param last_updated_by: The last_updated_by of this Scheduler.  # noqa: E501
        :type: str
        """

        self._last_updated_by = last_updated_by

    @property
    def created_by(self):
        """Gets the created_by of this Scheduler.  # noqa: E501


        :return: The created_by of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this Scheduler.


        :param created_by: The created_by of this Scheduler.  # noqa: E501
        :type: str
        """

        self._created_by = created_by

    @property
    def created_at(self):
        """Gets the created_at of this Scheduler.  # noqa: E501


        :return: The created_at of this Scheduler.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Scheduler.


        :param created_at: The created_at of this Scheduler.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this Scheduler.  # noqa: E501


        :return: The updated_at of this Scheduler.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this Scheduler.


        :param updated_at: The updated_at of this Scheduler.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    @property
    def archived_at(self):
        """Gets the archived_at of this Scheduler.  # noqa: E501


        :return: The archived_at of this Scheduler.  # noqa: E501
        :rtype: datetime
        """
        return self._archived_at

    @archived_at.setter
    def archived_at(self, archived_at):
        """Sets the archived_at of this Scheduler.


        :param archived_at: The archived_at of this Scheduler.  # noqa: E501
        :type: datetime
        """

        self._archived_at = archived_at

    @property
    def version(self):
        """Gets the version of this Scheduler.  # noqa: E501


        :return: The version of this Scheduler.  # noqa: E501
        :rtype: float
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this Scheduler.


        :param version: The version of this Scheduler.  # noqa: E501
        :type: float
        """

        self._version = version

    @property
    def type(self):
        """Gets the type of this Scheduler.  # noqa: E501


        :return: The type of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this Scheduler.


        :param type: The type of this Scheduler.  # noqa: E501
        :type: str
        """
        if type is None:
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501
        allowed_values = ["Cron", "OneTime"]  # noqa: E501
        if type not in allowed_values:
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    @property
    def workflow_id(self):
        """Gets the workflow_id of this Scheduler.  # noqa: E501


        :return: The workflow_id of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._workflow_id

    @workflow_id.setter
    def workflow_id(self, workflow_id):
        """Sets the workflow_id of this Scheduler.


        :param workflow_id: The workflow_id of this Scheduler.  # noqa: E501
        :type: str
        """
        if workflow_id is None:
            raise ValueError("Invalid value for `workflow_id`, must not be `None`")  # noqa: E501

        self._workflow_id = workflow_id

    @property
    def schedule(self):
        """Gets the schedule of this Scheduler.  # noqa: E501


        :return: The schedule of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._schedule

    @schedule.setter
    def schedule(self, schedule):
        """Sets the schedule of this Scheduler.


        :param schedule: The schedule of this Scheduler.  # noqa: E501
        :type: str
        """
        if schedule is None:
            raise ValueError("Invalid value for `schedule`, must not be `None`")  # noqa: E501

        self._schedule = schedule

    @property
    def parameters(self):
        """Gets the parameters of this Scheduler.  # noqa: E501


        :return: The parameters of this Scheduler.  # noqa: E501
        :rtype: WorkflowTemplateParameters
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """Sets the parameters of this Scheduler.


        :param parameters: The parameters of this Scheduler.  # noqa: E501
        :type: WorkflowTemplateParameters
        """

        self._parameters = parameters

    @property
    def organization_id(self):
        """Gets the organization_id of this Scheduler.  # noqa: E501


        :return: The organization_id of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this Scheduler.


        :param organization_id: The organization_id of this Scheduler.  # noqa: E501
        :type: str
        """
        if organization_id is None:
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def job_key(self):
        """Gets the job_key of this Scheduler.  # noqa: E501


        :return: The job_key of this Scheduler.  # noqa: E501
        :rtype: str
        """
        return self._job_key

    @job_key.setter
    def job_key(self, job_key):
        """Sets the job_key of this Scheduler.


        :param job_key: The job_key of this Scheduler.  # noqa: E501
        :type: str
        """

        self._job_key = job_key

    @property
    def last_run_at(self):
        """Gets the last_run_at of this Scheduler.  # noqa: E501


        :return: The last_run_at of this Scheduler.  # noqa: E501
        :rtype: datetime
        """
        return self._last_run_at

    @last_run_at.setter
    def last_run_at(self, last_run_at):
        """Sets the last_run_at of this Scheduler.


        :param last_run_at: The last_run_at of this Scheduler.  # noqa: E501
        :type: datetime
        """

        self._last_run_at = last_run_at

    @property
    def next_run_at(self):
        """Gets the next_run_at of this Scheduler.  # noqa: E501


        :return: The next_run_at of this Scheduler.  # noqa: E501
        :rtype: datetime
        """
        return self._next_run_at

    @next_run_at.setter
    def next_run_at(self, next_run_at):
        """Sets the next_run_at of this Scheduler.


        :param next_run_at: The next_run_at of this Scheduler.  # noqa: E501
        :type: datetime
        """

        self._next_run_at = next_run_at

    @property
    def is_paused(self):
        """Gets the is_paused of this Scheduler.  # noqa: E501


        :return: The is_paused of this Scheduler.  # noqa: E501
        :rtype: object
        """
        return self._is_paused

    @is_paused.setter
    def is_paused(self, is_paused):
        """Sets the is_paused of this Scheduler.


        :param is_paused: The is_paused of this Scheduler.  # noqa: E501
        :type: object
        """
        if is_paused is None:
            raise ValueError("Invalid value for `is_paused`, must not be `None`")  # noqa: E501

        self._is_paused = is_paused

    @property
    def organization(self):
        """Gets the organization of this Scheduler.  # noqa: E501


        :return: The organization of this Scheduler.  # noqa: E501
        :rtype: Organization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this Scheduler.


        :param organization: The organization of this Scheduler.  # noqa: E501
        :type: Organization
        """

        self._organization = organization

    @property
    def workflow(self):
        """Gets the workflow of this Scheduler.  # noqa: E501


        :return: The workflow of this Scheduler.  # noqa: E501
        :rtype: Workflow
        """
        return self._workflow

    @workflow.setter
    def workflow(self, workflow):
        """Sets the workflow of this Scheduler.


        :param workflow: The workflow of this Scheduler.  # noqa: E501
        :type: Workflow
        """

        self._workflow = workflow

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Scheduler, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Scheduler):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import functools
import inspect
from collections import defaultdict
from types import MethodType
from typing import List, Dict, Callable, Tuple, Union
from unittest.mock import MagicMock

from curia.api.swagger_client import PlatformApi
import curia.api.swagger_client.models as models_package
import re
from pymaybe import maybe
import uuid
from inspect import getframeinfo, stack

from curia.mock.server import Exception404
# pylint: disable=W0622,W0613
from curia.utils.string import to_snake_case, to_camel_case


def debug(message):
    caller = getframeinfo(stack()[1][0])
    print(f"[MockApi DEBUG: {caller.filename}:{caller.lineno}]: {message}")


class _MockApiInstanceBase:
    def __init__(self):
        seed_data, method_overrides = MockApiConfiguration.get_configuration()
        self.db = defaultdict(dict)
        self._load_seed_data(seed_data)
        self.method_overrides = method_overrides

    def _load_seed_data(self, seed_data):
        """
        Loads a list of seed models into the database

        :param seed_data: List of API models to load
        """
        for model in seed_data:
            self._load_model(model)

    def _load_model(self, model):
        self.db[type(model)][model.id] = model
        model_snake_case = to_snake_case(type(model).__name__)
        for key in model.attribute_map.keys():
            value = getattr(model, key)

            if type(value) in MODEL_MAP.values() and f"{key}_id" in model.attribute_map and "id" in value.attribute_map:
                # joined model case
                setattr(model, f"{key}_id", value.id)
                if model_snake_case in value.attribute_map:
                    setattr(value, model_snake_case, model)
                elif f"{model_snake_case}s" in value.attribute_map:
                    if isinstance(getattr(value, f"{model_snake_case}s"), list):
                        getattr(value, f"{model_snake_case}s").append(model)
                    else:
                        setattr(value, f"{model_snake_case}s", [model])
                self._load_model(value)
            if isinstance(value, list) and len(value) > 0 \
                    and hasattr(value[0], 'attribute_map') \
                    and f"{model_snake_case}_id" in value[0].attribute_map:
                # joined model case
                for joined_model in value:
                    setattr(joined_model, f"{model_snake_case}_id", model.id)

    def __getattribute__(self, name):
        try:
            method_overrides = object.__getattribute__(self, "method_overrides")
        except AttributeError:
            # handle case where object is still not fully initialized
            method_overrides = {}
        if name in method_overrides:
            return MethodType(method_overrides[name], self)
        return object.__getattribute__(self, name)

    def resolve_joins(self, models, joins: List[str]):
        assert len(models) > 0, "At least one model must be provided"
        for join in joins:
            current_models = models
            current_model_type = type(current_models[0])
            segmentation = join.split(".")
            for segment in segmentation:
                inverted_attribute_map = {value: key for key, value in current_model_type.attribute_map.items()}

                if segment not in inverted_attribute_map:
                    raise AttributeError(f"Cannot resolve join: {join}, {current_model_type} does not have an "
                                         f"attribute corresponding to {segment} (looking for '{segment}')")
                attribute_name = inverted_attribute_map[segment]
                attribute_swagger_type: str = current_model_type.swagger_types[attribute_name]
                if attribute_swagger_type.startswith("list["):
                    attribute_swagger_type = attribute_swagger_type[len("list["):-1]
                    next_model_type = MODEL_MAP[attribute_swagger_type]
                    current_model_type_snake_case = to_snake_case(current_model_type.__name__)
                    if f"{current_model_type_snake_case}_id" not in next_model_type.attribute_map:
                        raise ValueError(f"Cannot resolve join: {join}, could not resolve relationship between "
                                         f"{current_model_type} and type {next_model_type}. It's possible that this is a "
                                         f"many to many relationship, which is not yet supported by this mocking "
                                         f"framework, or it is possible that the relationship does not exist. In order "
                                         f"to resolve the relationship, {next_model_type} must have the attribute "
                                         f"{current_model_type_snake_case}_id so that the reverse relationship can be "
                                         f"established.")
                    next_models = []
                    for individual_model in current_models:
                        associated_models = [
                            value for value in self.db[next_model_type].values()
                            if getattr(value, f"{current_model_type_snake_case}_id") == individual_model.id
                        ]
                        setattr(individual_model, attribute_name, associated_models)
                        next_models.extend(associated_models)
                else:
                    next_model_type = MODEL_MAP[attribute_swagger_type]
                    next_models = []
                    for individual_model in current_models:
                        next_model_id = getattr(individual_model, f"{attribute_name}_id")
                        if next_model_id is None:
                            raise ValueError(f"{attribute_name}_id value on {current_models} is null!")
                        # get next model and set attribute of current model
                        next_model = self.db[next_model_type][next_model_id]
                        setattr(individual_model, attribute_name, next_model)
                        next_models.append(next_model)
                current_models = next_models
                current_model_type = next_model_type


def configure_mock_api(seed_data: List[object], method_overrides: Dict[str, Callable]):
    mock_api_configuration = MockApiConfiguration(seed_data=seed_data, method_overrides=method_overrides)

    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            with mock_api_configuration:
                return fn(*args, **kwargs)

        return wrapper

    return decorator


class MockApiConfiguration:
    STACK: List['MockApiConfiguration'] = []

    def __init__(self, seed_data: List[object], method_overrides: Dict[str, Callable]):
        """
        Initialize and __enter__ a MockApiConfiguration to seed the mock API with certain values. MockApiConfiguration
        contexts stack -- e.g.:

        with MockApiConfiguration(..., ...):
            with MockApiConfiguration(..., ...):
                api_instance = MockApiInstance()

        Will have the api_instance configured with the seed data from both configurations and the methods from both
        configurations, with the innermost context having precedence.

        :param seed_data: List of API models to seed into the mock database
        :param method_overrides: Dict mapping strings (the names of methods) to callables that implement those methods.
            The callables should expect to receive self as the first argument
        """
        self.seed_data = seed_data
        self.method_overrides = method_overrides

    def __enter__(self):
        MockApiConfiguration.STACK.append(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        MockApiConfiguration.STACK.pop()

    @classmethod
    def get_configuration(cls) -> Tuple[List[object], Dict[str, Callable]]:
        """
        Retrieves the current seed data and method overrides to instantiate a mock API instance

        :return: seed_data as concatenated seed_datas from all configs in STACK, method_overrides as combined
            dictionary of all method_overrides in STACK, with innermost configs taking precedence
        """
        seed_data = []
        method_overrides = {}
        for configuration in cls.STACK:
            seed_data.extend(configuration.seed_data)
            method_overrides.update(configuration.method_overrides)
        return seed_data, method_overrides


def get_models():
    """
    Gets all model classes in autogenerated API by stepping through the package

    :return: List of all model classes in autogenerated API
    """
    return {
        model_name: getattr(models_package, model_name)
        for model_name in dir(models_package)
        if isinstance(getattr(models_package, model_name), type)
    }


MODEL_MAP = get_models()


def fn_get_one(models, model_type):
    """
    Constructs a method for getting a single object from the API mock database for an arbitrary model_type
    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets a single model from the API mock database
    """

    def get_one(self, id, **kwargs):
        if id in self.db[model_type]:
            if kwargs.get("join", None) is not None:
                self.resolve_joins([self.db[model_type][id]], kwargs["join"])
            return self.db[model_type][id]
        raise Exception404("404 Model Not Found")

    return get_one


def fn_get_many(models, model_type):
    """
    Constructs a method for getting many objects from the API mock database for an arbitrary model_type
    See docs to implement more filter functionality:
    https://github.com/nestjsx/crud/wiki/Requests#filter

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets several models from the API mock database
    """

    def get_many(self, **kwargs):
        to_return = []
        for model in self.db[model_type].values():
            if kwargs.get("filter", None) is not None:
                filter = kwargs["filter"]
                for f in filter:
                    field, condition, value = f.split("||")
                    field_snake_case = to_snake_case(field)
                    try:
                        model_value = getattr(model, field_snake_case)
                    except AttributeError as e:
                        raise AttributeError(f"Model {model} does not have attribute {field_snake_case}!") from e
                    model_value_type = type(model_value)
                    # attempt to coerce type
                    try:
                        model_value = model_value_type(model_value)
                    except Exception as e:
                        debug(f"Failed to coerce type while processing filters; {e}. ")
                    valid = evaluate_condition(condition, model_value, value)
                    if not valid:
                        break
                else:
                    # for else: https://book.pythontips.com/en/latest/for_-_else.html
                    # only runs if there were no breaks (so no filters were invalid)
                    to_return.append(model)
        if "join" in kwargs:
            self.resolve_joins(to_return, kwargs["join"])
        return_type = models[f"GetMany{model_type.__name__}ResponseDto"]
        return return_type(data=to_return, count=len(to_return), total=len(to_return), page=0, page_count=1)

    return get_many


def evaluate_condition(condition, model_value, value):
    if condition == "$eq":
        valid = model_value == value
    elif condition == "$ne":
        valid = model_value != value
    elif condition == "$gt":
        valid = model_value > value
    elif condition == "$lt":
        valid = model_value < value
    elif condition == "$gte":
        valid = model_value >= value
    elif condition == "lte":
        valid = model_value <= value
    else:
        raise ValueError(f"Condition {condition} not recognized by the mock API. "
                         f"If you need it, implement it. (ask WK for reference as to how to do so)")
    return valid


def convert_model_to_model_cls(model: Union[dict, object], cls):
    """
    Converts a model, potentially as a dict,  to a model class instance. If the model is an object implementing to_dict,
    it will be converted to a dict first and then to the model class.
    """
    if hasattr(model, "to_dict"):
        model = model.to_dict()
    obj = cls(**model)
    # The object is now a model class instance and has swagger_types, so we can use those to recursively search
    # for nested models and convert them to model class instances
    for attr, attr_type in obj.swagger_types.items():
        if attr_type in MODEL_MAP and getattr(obj, attr) is not None:
            # Recursively convert nested models
            setattr(obj, attr, convert_model_to_model_cls(getattr(obj, attr), MODEL_MAP[attr_type]))
        # handle the case when attr_type represents a list of models
        elif attr_type.startswith("list[") and attr_type[5:-1] in MODEL_MAP and getattr(obj, attr) is not None:
            # Recursively convert nested models
            setattr(
                obj,
                attr,
                [convert_model_to_model_cls(item, MODEL_MAP[attr_type[5:-1]]) for item in getattr(obj, attr)]
            )
    return obj


def fn_create_one(models, model_type):
    """
    Constructs a method for creating one object the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that creates a object in the API mock database
    """

    def create_one(self, body, **kwargs):
        body = convert_model_to_model_cls(body, model_type)
        body.id = str(uuid.uuid4())
        self.db[model_type][body.id] = body
        return body

    return create_one


def fn_create_many(models, model_type):
    """
    Constructs a method for creating many objects in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that creates many objects in the API mock database
    """

    def create_many(self, body, **kwargs):
        if hasattr(body, "bulk"):
            bulk = body.bulk
        else:
            bulk = body["bulk"]
        models = []
        for model in bulk:
            model = convert_model_to_model_cls(model, model_type)
            model.id = str(uuid.uuid4())
            self.db[model_type][model.id] = model
            models.append(model)
        return models

    return create_many


def fn_update_one(models, model_type):
    """
    Constructs a method for updating one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that updates one object in the API mock database
    """

    def update_one(self, id, body, **kwargs):
        if hasattr(body, "to_dict"):
            body = body.to_dict()
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        result = self.db[model_type][id]
        for field, value in body.items():
            setattr(result, field, value)
        return result

    return update_one


def fn_delete_one(models, model_type):
    """
    Constructs a method for deleting one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that deletes one object in the API mock database
    """

    def delete_one(self, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        del self.db[model_type][id]
        return {"id": id, "deleted": True}

    return delete_one


def fn_replace_one(models, model_type):
    """
    Constructs a method for replacing one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that replaces one object in the API mock database
    """

    def replace_one(self, body, **kwargs):
        body = convert_model_to_model_cls(body, model_type)
        self.db[model_type][body.id] = body
        return body

    return replace_one


def fn_start(models, model_type):
    """
    Constructs a method for "starting" one object in the API mock database for an arbitrary model_type. No code
    is actually executed when the object is started, the status is just changed.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that starts one object in the API mock database
    """

    def start(self, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        self.db[model_type][id].status = "RUNNING"
        return self.db[model_type][id]

    return start


def fn_stop(models, model_type):
    """
    Constructs a method for "stopping" one object in the API mock database for an arbitrary model_type. No code
    is actually executed when the object is stopped, the status is just changed.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that stops one object in the API mock database
    """

    def stop(self, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        self.db[model_type][id].status = "ABORTED"
        return self.db[model_type][id]

    return stop


def fn_status(models, model_type):
    """
    Constructs a method for getting the status of one object in the API mock database for an arbitrary model_type.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets the status of one object in the API mock database
    """

    def status(self, id, **kwargs):
        if id in self.db[model_type]:
            if id not in self.db[model_type]:
                raise Exception404("404 Model Not Found")
            return self.db[model_type][id]
        raise Exception404("404 Model Not Found")

    return status


def default_functionality_builder(method, models):
    """
    Tries to parse the return type of a method from its docstring and creates a method that returns a MagicMock
    of that return type

    :param method: Method to mock
    :param models: List of models
    :return: function that mocks the default functionality of a model
    """
    return_line = [line.strip() for line in maybe(method.__doc__).split("\n").or_else("") if ":return:" in line]
    return_type = maybe(return_line or None).split()[1].or_else(None)

    if return_type is not None and return_type in models:
        @functools.wraps(method)
        def function(self, *args, **kwargs):
            return MagicMock(models[return_type])
    else:
        if return_type is not None:
            debug(f"{return_type} of method {method} is not a model; reverting to default MagicMock")

        @functools.wraps(method)
        def function(self, *args, **kwargs):
            return MagicMock()
    return function


# matches model names to functionality builders
regex_to_functionality_builder = {
    r"get_one_base_([a-z_]*)_controller_\1": fn_get_one,
    r"get_many_base_([a-z_]*)_controller_\1": fn_get_many,
    r"create_one_base_([a-z_]*)_controller_\1": fn_create_one,
    r"create_many_base_([a-z_]*)_controller_\1": fn_create_many,
    r"update_one_base_([a-z_]*)_controller_\1": fn_update_one,
    r"delete_one_base_([a-z_]*)_controller_\1": fn_delete_one,
    r"replace_one_base_([a-z_]*)_controller_\1": fn_replace_one,
    r"([a-z_]*)_controller_start": fn_start,
    r"([a-z_]*)_controller_stop": fn_stop,
    r"([a-z_]*)_controller_status": fn_status,
}


def get_functionality(models, method_name, method):
    """
    Parses a method name and builds the appropriate functionality around it, e.g. methods that look like
    create_one_base_{}_controller_{} will actually create a {} in the mock database

    :param models: List of parsed models
    :param method_name: Name of method to implement
    :param method: Actual method to implement
    :return: boolean representing whether functionality was parsed and created, Method mocking functionality
    """
    for pattern, functionality_builder in regex_to_functionality_builder.items():
        compiled_pattern = re.compile(pattern)
        match = compiled_pattern.fullmatch(method_name)
        if match:
            model_type_snake_case = match.group(1)
            model_type_camel_case = to_camel_case(model_type_snake_case)
            if model_type_camel_case not in models:
                print(f"Could not find model for method {method_name}! "
                      f"(parsed snake case: {model_type_snake_case})"
                      f"(parsed camel case: {model_type_camel_case})")
                break
            return True, functionality_builder(models, models[model_type_camel_case])
    return False, default_functionality_builder(method, models)


def build_method_dict():
    """
    Builds a dictionary of methods on the API and mocked functionality for each method
    :return: Dictionary of methods that can be used to mock the API
    """
    method_dict = {}
    debug(f"Found {len(MODEL_MAP.keys())} models")
    n_specialized = 0
    for method_name, method in inspect.getmembers(PlatformApi, predicate=inspect.isfunction):
        if method_name.startswith("__"):
            continue
        specialized, new_method_functionality = get_functionality(MODEL_MAP, method_name, method)
        method_dict[method_name] = new_method_functionality
        if specialized:
            n_specialized += 1

    debug(f"Found {len(method_dict.keys())} methods, {n_specialized} of which have autogenerated functionality.")
    return method_dict


def get_mock_api_instance_class():
    """
    Builds the mock api instance class by reading the actual api instance
    :return: MockApiInstance class
    """
    method_dict = build_method_dict()

    return type("MockApiInstance", (_MockApiInstanceBase,), method_dict)


MockApiInstance = get_mock_api_instance_class()



from curia.mock.api import MockApiInstance


class MockSession:
    def __init__(self, api_token: str = None, host: str = None, debug: bool = False):
        self.api_token = api_token
        self.host = host
        self.debug = debug
        self.api_instance = MockApiInstance()




import re


def first_lower(s):
    """
    Convert first letter of a string to lowercase
    :param s: String to make Lowercase
    :return: String with first letter lowercase
    """
    if len(s) == 0:
        return s
    return s[0].lower() + s[1:]


def to_camel_case(snake_str):
    """
    Convert snake_case to UpperCamelCase
    :param snake_str:
    :return: string in upper camel case
    """
    # We capitalize the first letter of each component
    # with the 'title' method and join them together.
    return ''.join(x.title() for x in snake_str.split('_'))


def to_lower_camel_case(snake_str):
    """
    Convert snake_case to lowerCamelCase
    :param snake_str:
    :return: string in lower camel case
    """
    return first_lower(to_camel_case(snake_str))


# see https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
to_snake_case_pattern = re.compile(r'(?<!^)(?=[A-Z])')


def to_snake_case(camel_str):
    """
    Convert UpperCamelCase or camelCase to snake_case
    :param camel_str:
    :return: string in snake case
    """
    return to_snake_case_pattern.sub('_', camel_str).lower()
